# Generated by Django 4.2.10 on 2024-02-23 11:01

import django.contrib.postgres.indexes
import django.contrib.postgres.search
import django.db.models.deletion
import mptt.fields
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("wbfdm", "0012_instrumentprice_created_instrumentprice_modified"),
    ]

    operations = [
        migrations.AddField(
            model_name="instrument",
            name="is_investable_universe",
            field=models.BooleanField(default=False),
        ),
        migrations.CreateModel(
            name="InstrumentDailyStatistics",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("statistics", models.JSONField(default=dict)),
                ("performances", models.JSONField(default=dict)),
                ("updated", models.DateTimeField(auto_now=True)),
                (
                    "instrument",
                    models.OneToOneField(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="daily_statistics",
                        to="wbfdm.instrument",
                    ),
                ),
            ],
            options={
                "verbose_name": "Daily Instrument Statistic",
                "verbose_name_plural": "Daily Instrument Statistics",
            },
        ),
        migrations.AlterField(
            model_name="instrumentprice",
            name="volume_200d",
            field=models.FloatField(
                blank=True,
                help_text="The Average Volume of the Asset over the last 200 days from the price date of the Asset.",
                null=True,
                verbose_name="Average Volume (200 Days)",
            ),
        ),
        migrations.AlterField(
            model_name="instrumentdailystatistics",
            name="performances",
            field=models.JSONField(default=dict, encoder=django.core.serializers.json.DjangoJSONEncoder),
        ),
        migrations.AlterField(
            model_name="instrumentdailystatistics",
            name="statistics",
            field=models.JSONField(default=dict, encoder=django.core.serializers.json.DjangoJSONEncoder),
        ),
        migrations.AddConstraint(
            model_name="classification",
            constraint=models.UniqueConstraint(fields=("group", "code_aggregated"), name="unique_classification"),
        ),
        migrations.AddConstraint(
            model_name="instrumentlistthroughmodel",
            constraint=models.UniqueConstraint(
                fields=("instrument", "instrument_list"), name="unique_instrument_per_list"
            ),
        ),
        migrations.AddConstraint(
            model_name="instrumentprice",
            constraint=models.UniqueConstraint(fields=("instrument", "date", "calculated"), name="unique_price"),
        ),
        migrations.AddField(
            model_name="instrument",
            name="level",
            field=models.PositiveIntegerField(default=0, editable=False),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name="instrument",
            name="lft",
            field=models.PositiveIntegerField(default=0, editable=False),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name="instrument",
            name="rght",
            field=models.PositiveIntegerField(default=0, editable=False),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name="instrument",
            name="tree_id",
            field=models.PositiveIntegerField(db_index=True, default=0, editable=False),
            preserve_default=False,
        ),
        migrations.AlterField(
            model_name="instrument",
            name="parent",
            field=mptt.fields.TreeForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="children",
                to="wbfdm.instrument",
                verbose_name="Parent Instrument",
            ),
        ),
        migrations.AlterField(
            model_name="instrumentclassificationrelatedinstrument",
            name="related_instrument",
            field=models.ForeignKey(
                limit_choices_to=models.Q(("instrument_type__is_classifiable", True), ("level", 0)),
                on_delete=django.db.models.deletion.CASCADE,
                related_name="instrument_classification_related",
                to="wbfdm.instrument",
            ),
        ),
        migrations.AlterField(
            model_name="instrumentclassificationthroughmodel",
            name="instrument",
            field=models.ForeignKey(
                limit_choices_to=models.Q(("instrument_type__is_classifiable", True), ("level", 0)),
                on_delete=django.db.models.deletion.CASCADE,
                related_name="classifications_through",
                to="wbfdm.instrument",
            ),
        ),
        migrations.AlterField(
            model_name="instrumentclassificationthroughmodel",
            name="related_instruments",
            field=models.ManyToManyField(
                blank=True,
                limit_choices_to=models.Q(("instrument_type__is_classifiable", True), ("level", 0)),
                through="wbfdm.InstrumentClassificationRelatedInstrument",
                to="wbfdm.instrument",
            ),
        ),
        migrations.AddField(
            model_name="instrument",
            name="last_price_date",
            field=models.DateField(blank=True, help_text="Last Price Date", null=True, verbose_name="Last Price Date"),
        ),
        migrations.AlterField(
            model_name="instrument",
            name="is_investable_universe",
            field=models.BooleanField(
                default=False,
                help_text="If True, the instrument belongs to the investable universe",
                verbose_name="In Investable Universe",
            ),
        ),
        migrations.AlterField(
            model_name="instrumentprice",
            name="market_capitalization_consolidated",
            field=models.FloatField(
                blank=True,
                help_text="the consolidated market value of a company in local currency.",
                null=True,
                verbose_name="Market Capitalization (Consolidated)",
            ),
        ),
        migrations.AlterField(
            model_name="controversy",
            name="instrument",
            field=models.ForeignKey(
                limit_choices_to=models.Q(("level", 0)),
                on_delete=django.db.models.deletion.CASCADE,
                to="wbfdm.instrument",
            ),
        ),
        migrations.AddField(
            model_name="instrument",
            name="search_vector",
            field=django.contrib.postgres.search.SearchVectorField(null=True),
        ),
        migrations.AddField(
            model_name="instrument",
            name="trigram_search_vector",
            field=models.CharField(blank=True, max_length=1024, null=True),
        ),
        migrations.AddIndex(
            model_name="instrument",
            index=django.contrib.postgres.indexes.GinIndex(fields=["search_vector"], name="instrument_sv_gin_idx"),
        ),
        migrations.AddIndex(
            model_name="instrument",
            index=django.contrib.postgres.indexes.GinIndex(
                fields=["trigram_search_vector"], name="instrument_trigram_sv_gin_idx", opclasses=["gin_trgm_ops"]
            ),
        ),
        migrations.AddIndex(
            model_name="instrument",
            index=models.Index(fields=["parent", "exchange", "isin"], name="instrument_children_index"),
        ),
    ]



from decimal import Decimal
from typing import Any

import numpy as np
import pandas as pd
import plotly.graph_objects as go
from dateutil.relativedelta import relativedelta
from django.contrib.messages import warning
from django.db.models import Case, Exists, F, FloatField, OuterRef, When, Window
from django.db.models.functions import Lead
from rest_framework import filters
from wbcore import viewsets
from wbcore.contrib.io.viewsets import ExportPandasAPIViewSet
from wbcore.filters import DjangoFilterBackend
from wbcore.pandas import fields as pf
from wbcore.permissions.permissions import InternalUserPermissionMixin
from wbcore.utils.date import get_date_interval_from_request
from wbcore.utils.figures import (
    get_default_timeserie_figure,
    get_hovertemplate_timeserie,
)
from wbcore.utils.strings import format_number

from wbfdm.analysis.financial_analysis.financial_statistics_analysis import (
    FinancialStatistics,
)
from wbfdm.filters.instrument_prices import (
    InstrumentPriceFilterSet,
    InstrumentPriceFinancialStatisticsChartFilterSet,
    InstrumentPriceFrequencyFilter,
    InstrumentPriceInstrumentFilterSet,
    InstrumentPriceSingleBenchmarkFilterSet,
)
from wbfdm.import_export.resources.instrument_prices import (
    InstrumentPriceExportResource,
)
from wbfdm.models import Instrument, InstrumentPrice
from wbfdm.serializers import (
    InstrumentPriceInstrumentModelSerializer,
    InstrumentPriceModelSerializer,
)
from wbfdm.viewsets.configs import (
    BestAndWorstReturnsInstrumentEndpointConfig,
    BestAndWorstReturnsInstrumentPandasDisplayConfig,
    BestAndWorstReturnsInstrumentTitleConfig,
    FinancialStatisticsInstrumentButtonConfig,
    FinancialStatisticsInstrumentEndpointConfig,
    FinancialStatisticsInstrumentPandasDisplayConfig,
    FinancialStatisticsInstrumentTitleConfig,
    InstrumentPriceButtonConfig,
    InstrumentPriceDisplayConfig,
    InstrumentPriceInstrumentButtonConfig,
    InstrumentPriceInstrumentDistributionReturnsChartEndpointConfig,
    InstrumentPriceInstrumentDistributionReturnsChartTitleConfig,
    InstrumentPriceInstrumentEndpointConfig,
    InstrumentPriceInstrumentTitleConfig,
    InstrumentPriceStatisticsInstrumentEndpointConfig,
    InstrumentPriceStatisticsInstrumentTitleConfig,
    InstrumentPriceTitleConfig,
)

from ..mixins import InstrumentMixin


class InstrumentPriceModelViewSet(viewsets.ModelViewSet):
    IDENTIFIER = "wbfdm:price"
    filter_backends = (DjangoFilterBackend, filters.OrderingFilter)
    queryset = InstrumentPrice.objects.all()
    serializer_class = InstrumentPriceModelSerializer

    filterset_class = InstrumentPriceFilterSet
    ordering = ["-date"]
    ordering_fields = [
        "date",
        "gross_value",
        "daily_diff_gross_value",
        "net_value",
        "daily_diff_net_value",
        "market_capitalization",
        "sharpe_ratio",
        "correlation",
        "beta",
        "outstanding_shares_consolidated",
        "volume",
        "volume_50d",
    ]

    display_config_class = InstrumentPriceDisplayConfig
    title_config_class = InstrumentPriceTitleConfig
    button_config_class = InstrumentPriceButtonConfig

    def get_queryset(self):
        return (
            super()
            .get_queryset()
            .annotate(
                real_price_exists=Exists(
                    InstrumentPrice.objects.filter(
                        instrument=OuterRef("instrument_id"), calculated=False, date=OuterRef("date")
                    )
                ),
                currency_symbol=F("instrument__currency__symbol"),
            )
            .annotate_base_data()
            .select_related("instrument")
        )


class InstrumentPriceInstrumentModelViewSet(InstrumentMixin, InstrumentPriceModelViewSet):
    IDENTIFIER = "wbfdm:instrument-price"

    title_config_class = InstrumentPriceInstrumentTitleConfig
    button_config_class = InstrumentPriceInstrumentButtonConfig
    endpoint_config_class = InstrumentPriceInstrumentEndpointConfig

    filterset_class = InstrumentPriceInstrumentFilterSet
    serializer_class = InstrumentPriceInstrumentModelSerializer
    IMPORT_ALLOWED = False

    def get_resource_class(self):
        return InstrumentPriceExportResource

    def get_aggregates(self, queryset, paginated_queryset):
        res = dict()
        if queryset.exists():
            if (first_net_value := queryset.earliest("date").net_value) and (
                last_net_value := queryset.latest("date").net_value
            ):
                diff_net = last_net_value / first_net_value - 1 if first_net_value != 0 else 0
                res["daily_diff_net_value"] = {"∆": format_number(diff_net)}
            if (first_gross_value := queryset.earliest("date").gross_value) and (
                last_gross_value := queryset.latest("date").gross_value
            ):
                diff_gross = last_gross_value / first_gross_value - 1 if first_gross_value != 0 else 0
                res["daily_diff_gross_value"] = {"∆": format_number(diff_gross)}
        return res

    def get_queryset(self):
        queryset = super().get_queryset().filter(instrument=self.instrument)
        if "calculated" not in self.request.GET:
            queryset = queryset.filter_only_valid_prices()
        return (
            queryset.annotate(
                last_net_value=Window(Lead("net_value"), order_by="-date"),
                last_gross_value=Window(Lead("gross_value"), order_by="-date"),
                daily_diff_net_value=Case(
                    When(last_net_value=0, then=None),
                    default=F("net_value") / F("last_net_value") - 1,
                    output_field=FloatField(),
                ),
                daily_diff_gross_value=Case(
                    When(last_gross_value=0, then=None),
                    default=F("gross_value") / F("last_gross_value") - 1,
                    output_field=FloatField(),
                ),
            )
            .select_related("instrument")
            .select_related("import_source")
        )


class InstrumentPriceInstrumentStatisticsChartView(InstrumentMixin, viewsets.ChartViewSet):
    IDENTIFIER = "wbfdm:instrument-statisticschart"
    queryset = InstrumentPrice.objects.all()

    title_config_class = InstrumentPriceStatisticsInstrumentTitleConfig
    endpoint_config_class = InstrumentPriceStatisticsInstrumentEndpointConfig

    def get_queryset(self):
        return InstrumentPrice.objects.filter(instrument=self.instrument, calculated=False)

    def get_plotly(self, queryset):
        fig = get_default_timeserie_figure()

        if self.instrument.related_instruments.count() > 0:
            reference = self.instrument.related_instruments.first().name_repr
            risk = self.instrument.primary_risk_instrument.name_repr
            df = pd.DataFrame(queryset.values("date", "sharpe_ratio", "correlation", "beta")).replace(
                [np.inf, -np.inf], np.nan
            )

            if not df.empty:
                df = df.set_index("date").sort_index().dropna()
                fig.add_trace(
                    go.Scatter(
                        x=df.index,
                        y=df.sharpe_ratio,
                        mode="lines",
                        name=f"Sharpe Ratio ({risk})",
                        hovertemplate=get_hovertemplate_timeserie(currency=""),
                    )
                )
                fig.add_trace(
                    go.Scatter(
                        x=df.index,
                        y=df.correlation,
                        mode="lines",
                        name=f"Correlation ({reference})",
                        hovertemplate=get_hovertemplate_timeserie(currency=""),
                    )
                )
                fig.add_trace(
                    go.Scatter(
                        x=df.index,
                        y=df.beta,
                        mode="lines",
                        name=f"Beta ({reference})",
                        hovertemplate=get_hovertemplate_timeserie(currency=""),
                    )
                )
        else:
            warning(
                self.request,
                "The chart is empty because there is no benchmark or risk free rate associated to this instrument",
                extra_tags="auto_close=0",
            )

        return fig


class FinancialStatisticsInstrumentPandasView(InstrumentMixin, InternalUserPermissionMixin, ExportPandasAPIViewSet):
    IDENTIFIER = "wbfdm:financialstatistics"

    LIST_DOCUMENTATION = "wbfdm/markdown/documentation/financial_statistics.md"

    queryset = InstrumentPrice.objects.all()
    endpoint_config_class = FinancialStatisticsInstrumentEndpointConfig
    title_config_class = FinancialStatisticsInstrumentTitleConfig
    display_config_class = FinancialStatisticsInstrumentPandasDisplayConfig
    button_config_class = FinancialStatisticsInstrumentButtonConfig
    pandas_fields = pf.PandasFields(
        fields=(
            pf.PKField(key="id", label="ID"),
            pf.CharField(key="financial", label="Financial"),
            pf.CharField(key="instrument_statistics", label="Instrument"),
            pf.CharField(key="benchmark_statistics", label="Benchmark"),
            pf.CharField(key="instrument_one_year", label="Instrument - One Year"),
            pf.CharField(key="benchmark_one_year", label="Benchmark - One Year"),
            pf.BooleanField(key="instrument_vs_benchmark", label="Instrument VS Benchmark"),
            pf.BooleanField(key="instrument_vs_benchmark_one_year", label="Instrument VS Benchmark - One Year"),
        )
    )

    filterset_class = InstrumentPriceSingleBenchmarkFilterSet
    search_fields = ("financial",)
    ordering_fields = (
        "financial" "instrument_statistics",
        "benchmark_statistics",
        "instrument_one_year",
        "benchmark_one_year",
    )

    def get_dataframe(self, request, queryset, **kwargs):
        starting_date, end_date = get_date_interval_from_request(request, exclude_weekend=True)

        if request.GET.get("benchmark", None):
            benchmark = Instrument.objects.get(id=request.GET["benchmark"])
        else:
            benchmark = self.instrument.primary_benchmark

        df = pd.DataFrame()

        instrument_prices = self.instrument.get_prices_df(from_date=starting_date, to_date=end_date)
        instrument_risk_free_rate_prices_df = (
            self.instrument.primary_risk_instrument.get_prices_df(from_date=starting_date, to_date=end_date)
            if self.instrument.primary_risk_instrument
            else pd.Series(dtype=float)
        )

        benchmark_prices = (
            benchmark.get_prices_df(from_date=starting_date, to_date=end_date) if benchmark else pd.Series(dtype=float)
        )
        benchmark_risk_free_rate_prices_df = (
            benchmark.primary_risk_instrument.get_prices_df(from_date=starting_date, to_date=end_date)
            if (benchmark and benchmark.primary_risk_instrument)
            else pd.Series(dtype=float)
        )

        if instrument_prices.empty or instrument_prices.shape[0] < 2:
            return df

        df["financial"] = [
            "Name",
            "Date From",
            "To Date",
            "Daily Mean Return",
            "Compound Annual Growth Rate",
            "Last Cumulative Return",
            "Ann. Volatility",
            "Beta",
            "Correlation",
            "Risk free rate",
            "Sharpe Ratio",
            "Maximum Drawdown",
            "Maximum Drawdown Date",
            "Longest Drawdown Period",
            "Last Maximum Drawdown",
            "Last Maximum Drawdown Date",
            "Value at Risk 1%",
            "Value at Risk 5%",
            "Value at Risk 10%",
            "Conditional Value at Risk 1%",
            "Conditional Value at Risk 5%",
            "Conditional Value at Risk 10%",
            "Skewness",
            "Excess Kurtosis",
            "Sortino Ratio",
            "Adj. Sortino Ratio",
            "Calmar Ratio",
            "Sterling Ratio",
            "Burke Ratio",
        ]

        def metric_string_adjusted(metric: [float, None], in_pct: bool = True):
            str_adj = ""
            if metric:
                str_adj = f"{round(metric * 100, 2)}%" if in_pct else f"{round(metric, 2)}"
            return str_adj

        def fill_df(
            _instrument_title: str,
            _instrument_statistics: FinancialStatistics,
            _benchmark_prices_df: pd.DataFrame,
            _risk_free_rate_prices_df: pd.DataFrame,
        ) -> list[Any]:
            return [
                _instrument_title,
                _instrument_statistics.start.strftime("%Y-%m-%d"),
                _instrument_statistics.end.strftime("%Y-%m-%d"),
                metric_string_adjusted(_instrument_statistics.get_mean_return()),
                metric_string_adjusted(_instrument_statistics.get_compound_annual_growth_rate()),
                metric_string_adjusted(_instrument_statistics.get_last_cumulative_return()),
                metric_string_adjusted(_instrument_statistics.get_volatility()),
                metric_string_adjusted(_instrument_statistics.get_beta(benchmark_prices_df=_benchmark_prices_df)),
                metric_string_adjusted(
                    _instrument_statistics.get_correlation(benchmark_prices_df=_benchmark_prices_df)
                ),
                metric_string_adjusted(_instrument_statistics.get_risk_free_rate(_risk_free_rate_prices_df)),
                metric_string_adjusted(
                    _instrument_statistics.get_sharpe_ratio(_risk_free_rate_prices_df), in_pct=False
                ),
                metric_string_adjusted(_instrument_statistics.get_maximum_drawdown()),
                _instrument_statistics.get_maximum_drawdown_date(),
                f"{_instrument_statistics.get_longest_drawdown_period()} days",
                metric_string_adjusted(_instrument_statistics.get_last_recent_maximum_drawdown()),
                _instrument_statistics.get_last_recent_maximum_drawdown_date(),
                metric_string_adjusted(_instrument_statistics.get_value_at_risk(alpha=0.01)),
                metric_string_adjusted(_instrument_statistics.get_value_at_risk(alpha=0.05)),
                metric_string_adjusted(_instrument_statistics.get_value_at_risk(alpha=0.1)),
                metric_string_adjusted(_instrument_statistics.get_conditional_value_at_risk(alpha=0.01)),
                metric_string_adjusted(_instrument_statistics.get_conditional_value_at_risk(alpha=0.05)),
                metric_string_adjusted(_instrument_statistics.get_conditional_value_at_risk(alpha=0.10)),
                metric_string_adjusted(_instrument_statistics.get_skewness(), in_pct=False),
                metric_string_adjusted(_instrument_statistics.get_excess_kurtosis(), in_pct=False),
                metric_string_adjusted(_instrument_statistics.get_sortino_ratio(), in_pct=False),
                metric_string_adjusted(_instrument_statistics.get_adjusted_sortino_ratio(), in_pct=False),
                metric_string_adjusted(_instrument_statistics.get_calmar_ratio(), in_pct=False),
                metric_string_adjusted(
                    _instrument_statistics.get_sterling_ratio(_risk_free_rate_prices_df), in_pct=False
                ),
                metric_string_adjusted(
                    _instrument_statistics.get_burke_ratio(_risk_free_rate_prices_df), in_pct=False
                ),
            ]

        df["instrument_statistics"] = fill_df(
            self.instrument.name,
            FinancialStatistics(instrument_prices),
            benchmark_prices,
            instrument_risk_free_rate_prices_df,
        )
        instrument_before_previous_year_date = max(
            instrument_prices.index[0], instrument_prices.index[-1] - relativedelta(years=1)
        )
        df["instrument_one_year"] = fill_df(
            self.instrument.name,
            FinancialStatistics(instrument_prices.truncate(before=instrument_before_previous_year_date)),
            benchmark_prices.truncate(before=instrument_before_previous_year_date),
            instrument_risk_free_rate_prices_df.truncate(before=instrument_before_previous_year_date),
        )

        if not benchmark_prices.empty:
            df["benchmark_statistics"] = fill_df(
                benchmark.name,
                FinancialStatistics(benchmark_prices),
                benchmark_prices,
                benchmark_risk_free_rate_prices_df,
            )
            benchmark_before_previous_year_date = max(
                benchmark_prices.index[0], benchmark_prices.index[-1] - relativedelta(years=1)
            )

            df["benchmark_one_year"] = fill_df(
                benchmark.name,
                FinancialStatistics(benchmark_prices.truncate(before=benchmark_before_previous_year_date)),
                benchmark_prices.truncate(before=benchmark_before_previous_year_date),
                benchmark_risk_free_rate_prices_df.truncate(before=benchmark_before_previous_year_date),
            )
            df["instrument_vs_benchmark"] = df[["instrument_statistics", "benchmark_statistics"]].apply(
                lambda x: x["instrument_statistics"] - x["benchmark_statistics"] > 0
                if type(x["instrument_statistics"]) in [float, Decimal]
                and type(x["benchmark_statistics"]) in [float, Decimal]
                else False,
                axis=1,
            )
            df["instrument_vs_benchmark_one_year"] = df[["instrument_one_year", "benchmark_one_year"]].apply(
                lambda x: x["instrument_one_year"] - x["benchmark_one_year"] > 0
                if type(x["instrument_one_year"]) in [float, Decimal]
                and type(x["benchmark_one_year"]) in [float, Decimal]
                else False,
                axis=1,
            )
        df["id"] = df.index
        return df

    def manipulate_dataframe(self, df):
        return df.where(pd.notnull(df), 0)


class InstrumentPriceInstrumentDistributionReturnsChartView(InstrumentMixin, viewsets.ChartViewSet):
    IDENTIFIER = "wbfdm:instrument-distributionreturnschart"
    queryset = InstrumentPrice.objects.all()

    title_config_class = InstrumentPriceInstrumentDistributionReturnsChartTitleConfig
    endpoint_config_class = InstrumentPriceInstrumentDistributionReturnsChartEndpointConfig
    filterset_class = InstrumentPriceFinancialStatisticsChartFilterSet

    def get_plotly(self, queryset):
        fig = go.Figure()

        def update_layout(text):
            fig.update_layout(
                title={"text": text, "y": 0, "x": 0.5, "xanchor": "center", "yanchor": "bottom"},
                yaxis_title="Density",
                xaxis_title="Returns in %",
                legend_title="Instruments",
            )

        if self.request.GET.get("benchmark", None):
            benchmark = Instrument.objects.get(id=self.request.GET["benchmark"])
        else:
            benchmark = self.instrument.primary_benchmark

        starting_date, end_date = get_date_interval_from_request(self.request, exclude_weekend=True)
        first_day_available_of_instrument = self.instrument.inception_date.strftime("%Y-%m-%d")
        instrument_prices = self.instrument.get_prices_df(from_date=starting_date, to_date=end_date)

        if instrument_prices.empty:
            update_layout(f"{self.instrument} has no data or not enough data")
            return fig
        if not starting_date:
            starting_date = instrument_prices.index[0]
        if not end_date:
            end_date = instrument_prices.index[-1]

        returns_df = pd.DataFrame()
        returns_df[self.instrument] = (
            self.instrument._compute_performance(instrument_prices, freq=self.request.GET.get("frequency"))[
                "performance"
            ]
            * 100
        )
        returns_df.columns = [f"{self.instrument} - #data:{len(returns_df[self.instrument].dropna())}"]
        import plotly.figure_factory as ff

        if returns_df.empty:
            return go.Figure()
        returns_df = returns_df.astype(float)
        fig = ff.create_distplot(
            [returns_df[c] for c in returns_df.columns],
            [f"{returns_df.columns[x]}" for x in range(len(returns_df.columns))],
            bin_size=0.3,
        )
        if not benchmark:
            update_layout(f"Inception Date -- {self.instrument.name}: {first_day_available_of_instrument}")
            return fig

        benchmark_prices = benchmark.get_prices_df(from_date=starting_date, to_date=end_date)
        if benchmark_prices.empty:
            update_layout(
                f"Inception Date -- {self.instrument.name}: {first_day_available_of_instrument}"
                f" // {benchmark} has no data"
            )
            return fig

        if starting_date < benchmark.inception_date:
            update_layout(
                f"Inception Date -- {self.instrument.name}: {first_day_available_of_instrument}"
                f" // {benchmark.name}: {benchmark.inception_date.strftime('%Y-%m-%d')}"
            )
            return fig
        returns_df[benchmark] = benchmark.extract_daily_performance_df(benchmark_prices)["performance"] * 100
        returns_df.columns = [
            returns_df.columns[0],
            f"{benchmark} - #data:{len(returns_df[benchmark].dropna())}",
        ]
        returns_df = returns_df.dropna()

        fig = ff.create_distplot(
            [returns_df[c] for c in returns_df.columns],
            [f"{returns_df.columns[x]}" for x in range(len(returns_df.columns))],
            bin_size=0.3,
        )
        update_layout(
            f"Inception Date -- {self.instrument.name}: {first_day_available_of_instrument}"
            f" // {benchmark.name}: {benchmark.inception_date.strftime('%Y-%m-%d')}"
        )
        return fig


class BestAndWorstReturnsInstrumentPandasView(InstrumentMixin, InternalUserPermissionMixin, ExportPandasAPIViewSet):
    IDENTIFIER = "wbfdm:bestandworstreturns"

    queryset = InstrumentPrice.objects.all()
    endpoint_config_class = BestAndWorstReturnsInstrumentEndpointConfig
    title_config_class = BestAndWorstReturnsInstrumentTitleConfig
    display_config_class = BestAndWorstReturnsInstrumentPandasDisplayConfig

    pandas_fields = pf.PandasFields(
        fields=(
            pf.PKField(key="id", label="ID"),
            pf.DateField(key="date_best_returns", label="Date Best Returns"),
            pf.FloatField(key="best_returns", label="Best Returns", precision=3, percent=True),
            pf.DateField(key="date_worst_returns", label="Date Worst Returns"),
            pf.FloatField(key="worst_returns", label="Worst Returns", precision=3, percent=True),
        )
    )

    filterset_class = InstrumentPriceFrequencyFilter
    ordering_fields = ("date_best_returns" "best_returns", "date_worst_returns", "worst_returns")

    def get_dataframe(self, request, queryset, **kwargs):
        df_to_display = pd.DataFrame()
        from_date, to_date = get_date_interval_from_request(self.request, date_range_fieldname="date")
        if not (prices_df := self.instrument.get_prices_df(from_date=from_date, to_date=to_date)).empty:
            returns_df = FinancialStatistics(prices_df).get_best_and_worst_returns(freq=request.GET.get("frequency"))
            if not returns_df.empty and returns_df.shape[0] > 0:
                df_to_display["date_best_returns"] = returns_df["Date Best Return"].dt.strftime("%Y-%m-%d")
                df_to_display["best_returns"] = returns_df["Best Return"]
                df_to_display["date_worst_returns"] = returns_df["Date Worst Return"].dt.strftime("%Y-%m-%d")
                df_to_display["worst_returns"] = returns_df["Worst Return"]

            df_to_display["id"] = df_to_display.index
        return df_to_display



from django.utils.functional import cached_property
from wbcore import viewsets

from wbfdm.models.instruments import InstrumentRequest
from wbfdm.serializers.instruments.instrument_requests import (
    InstrumentRequestModelSerializer,
    InstrumentRequestRepresentationSerializer,
)

from ..configs import (
    InstrumentRequestDisplayConfig,
    InstrumentRequestEndpointConfig,
    InstrumentRequestTitleConfig,
)


class InstrumentRequestRepresentationViewSet(viewsets.RepresentationViewSet):
    queryset = InstrumentRequest.objects.all()
    serializer_class = InstrumentRequestRepresentationSerializer
    search_fields = ("notes", "requester__computed_str")
    ordering_fields = ["created"]
    ordering = ["-created"]


class InstrumentRequestModelViewSet(viewsets.ModelViewSet):
    queryset = InstrumentRequest.objects.select_related("requester", "handler", "created_instrument")
    serializer_class = InstrumentRequestModelSerializer
    filterset_fields = {
        "status": ["exact"],
        "requester": ["exact"],
        "notes": ["icontains"],
        "created": ["gte", "exact", "lte"],
    }

    ordering_fields = ("status", "requester", "notes", "created")
    ordering = ("-created",)
    search_fields = ("notes", "requester__computed_str", "instrument_data__isin")

    display_config_class = InstrumentRequestDisplayConfig
    title_config_class = InstrumentRequestTitleConfig
    endpoint_config_class = InstrumentRequestEndpointConfig

    @cached_property
    def has_validation_permission(self):
        return self.request.user.has_perm("wbfdm.administrate_instrument")

    # def get_queryset(self):
    #     return (
    #         super(InstrumentRequestModelViewSet, self)
    #         .get_queryset()
    #         .annotate(**{field: F(field) for field in INSTRUMENT_BASE_FIELDS})
    #     )



from django.utils.translation import gettext as _
from wbcore.menus import ItemPermission, MenuItem
from wbcore.permissions.shortcuts import is_internal_user

INSTRUMENT_MENU_ITEM = MenuItem(
    label=_("Instruments"),
    endpoint="wbfdm:instrument-list",
    permission=ItemPermission(
        method=lambda request: is_internal_user(request.user), permissions=["wbfdm.view_instrument"]
    ),
)



import typing

from rest_framework.reverse import reverse
from wbcore.contrib.color.enums import WBColor
from wbcore.metadata.configs import display as dp
from wbcore.metadata.configs.display import DisplayViewConfig

if typing.TYPE_CHECKING:
    from wbfdm.viewsets import StatementWithEstimatesPandasViewSet


class StatementWithEstimatesDisplayViewConfig(DisplayViewConfig):
    view: "StatementWithEstimatesPandasViewSet"

    ESTIMATE_COLOR = "#D2E5F6"
    DEFAULT_COL_WIDTH = 100

    def get_list_display(self) -> dp.ListDisplay:
        def generate_formatting_rules(col_key: str) -> typing.Iterator[dp.FormattingRule]:
            yield dp.FormattingRule(
                style={"color": WBColor.RED_DARK.value, "fontWeight": "bold"},
                condition=[("<", 0), ("!=", "Net Debt", "financial"), ("!=", "Net Debt/Shares", "financial")],
            )
            yield dp.FormattingRule(
                condition=[("<", 0), ("==", "Net Debt", "financial")],
                style={"color": WBColor.GREEN_DARK.value, "fontWeight": "bold"},
            )
            yield dp.FormattingRule(
                condition=[("<", 0), ("==", "Net Debt/Shares", "financial")],
                style={"color": WBColor.GREEN_DARK.value, "fontWeight": "bold"},
            )
            yield dp.FormattingRule(
                condition=[(">", 0), ("==", "Net Debt", "financial")],
                style={"color": WBColor.RED_DARK.value, "fontWeight": "bold"},
            )
            yield dp.FormattingRule(
                condition=[(">", 0), ("==", "Net Debt/Shares", "financial")],
                style={"color": WBColor.RED_DARK.value, "fontWeight": "bold"},
            )
            if self.view.estimate_mapping[col_key] is True:
                yield dp.FormattingRule(style={"background-color": self.ESTIMATE_COLOR})

        def generate_year_field(year_col) -> dp.Field:
            year = year_col[:-2]
            year_column = dp.Field(
                key=year,
                label=year,
                open_by_default=year_col in self.view.year_columns and bool(self.view.df[year_col].isna().all()),
                children=[
                    *map(
                        lambda interim_col: dp.Field(
                            key=interim_col,
                            label=interim_col[5:],
                            show="open",
                            formatting_rules=generate_formatting_rules(interim_col),
                            width=self.DEFAULT_COL_WIDTH,
                        ),
                        filter(lambda col: year in col, self.view.interim_columns),
                    ),
                    dp.Field(
                        key=year_col,
                        label="Y",
                        formatting_rules=generate_formatting_rules((year_col)),
                        width=self.DEFAULT_COL_WIDTH,
                    ),
                ],
            )

            return year_column

        return dp.ListDisplay(
            fields=[
                dp.Field(key="progress", label="Yearly Trend", pinned="left"),
                *map(generate_year_field, self.view.year_columns),
            ],
            legends=[dp.Legend(items=[dp.LegendItem(label="Estimated", icon=self.ESTIMATE_COLOR)])],
            # formatting=[
            #     dp.Formatting(
            #         column="financial",
            #         formatting_rules=[
            #             dp.FormattingRule(
            #                 condition=("==", "Net Debt/Shares"),
            #                 style={"color": "red"},
            #             )
            #         ]
            #     )
            # ],
            tree=True,
            tree_group_field="financial",
            tree_group_label="Financial",
            tree_group_pinned="left",
            tree_group_level_options=[
                dp.TreeGroupLevelOption(
                    list_endpoint=reverse(
                        "wbfdm:financial_metric_analysis-list",
                        args=[self.view.kwargs.get("instrument_id")],
                        request=self.request,
                    ),
                )
            ],
        )



from wbcore.metadata.configs.titles import TitleViewConfig

from wbfdm.enums import MarketDataChartType


class PerformanceSummaryChartTitleConfig(TitleViewConfig):
    def get_list_title(self):
        return f"{self.view.instrument} - Performance Summary"


class MarketDataChartTitleConfig(TitleViewConfig):
    def get_list_title(self):
        market_data_chart_type = MarketDataChartType(self.request.GET.get("chart_type", "close"))
        return f"{self.view.instrument} - {market_data_chart_type.label} Chart"



class MetricInvalidParameterException(Exception):
    """
    If instantiated backend parameters are invalid, the orchestrator or backend itself are expected to raise that exception
    """

    pass
